import click
import pyperclip
from tabulate import tabulate

from kobodl import actions, cli
from kobodl.globals import Globals
from kobodl.kobo import Kobo
from kobodl.settings import User


@click.group(name='user', short_help='show and create users')
def user():
    pass


@user.command(name='list', help='list all users')
@click.pass_obj
def list(ctx):
    userlist = Globals.Settings.UserList.users
    headers = ['Email', 'UserKey', 'DeviceId']
    data = sorted(
        [
            (
                user.Email,
                user.UserKey,
                user.DeviceId,
            )
            for user in userlist
        ]
    )
    click.echo(tabulate(data, headers, tablefmt=ctx['fmt']))


@user.command(name='rm', help='remove user by Email, UserKey, or DeviceID')
@click.argument('identifier', type=click.STRING)
@click.pass_obj
def list(ctx, identifier):
    removed = Globals.Settings.UserList.removeUser(identifier)
    if removed:
        Globals.Settings.Save()
        click.echo(f'Removed {removed.Email}')
    else:
        click.echo(f'No user with email, key, or device id that matches "{identifier}"')


@user.command(name='add', help='add new user')
@click.option('--email', prompt=True, hide_input=False, type=click.STRING, help="kobo.com email.")
@click.password_option(help="kobo.com password (not stored)")
@click.pass_obj
def add(ctx, email, password):
    user = User(Email=email)
    click.echo(
        """
    1. Open https://authorize.kobo.com/signin in a private/incognito window in your browser.
    2. wait till the page loads (do not login!)
    3. open the developer tools (use F12 in Firefox/Chrome, or right-click and choose "inspect")
    4. select the console tab,
    5. copy-paste the following code to the console there and then press Enter.

    var newCaptchaDiv = document.createElement("div");
    newCaptchaDiv.id = "new-hcaptcha-container";
    var siteKey = document.getElementById('hcaptcha-container').getAttribute('data-sitekey');
    document.body.replaceChildren(newCaptchaDiv);
    grecaptcha.render(newCaptchaDiv.id, {
        sitekey: siteKey,
        callback: function(r) {console.log("Captcha response:");console.log(r);}
    });
    console.log('Click the checkbox to get the code');

    A captcha should show up below the Sign-in form. Once you solve the captcha its response will be written
    below the pasted code in the browser's console. Copy the response (the line below "Captcha response:")
    and paste it here.  It will be very long!
    """
    )
    input('Press enter after copying the captcha code...')
    captcha = pyperclip.paste().strip()
    click.echo(f'Read captcha code from clipboard: {captcha}')
    actions.Login(user, password, captcha)
    Globals.Settings.UserList.users.append(user)
    Globals.Settings.Save()
    click.echo('Login Success. Try to list your books with `kobodl book list`')


cli.add_command(user)



import base64
import dataclasses
import html
import os
import re
import sys
import urllib
import uuid
from enum import Enum
from shutil import copyfile
from typing import Dict, Tuple

import requests
from bs4 import BeautifulSoup
from dataclasses_json import dataclass_json

from kobodl.debug import debug_data
from kobodl.globals import Globals
from kobodl.koboDrmRemover import KoboDrmRemover
from kobodl.settings import User


@dataclass_json
@dataclasses.dataclass
class Book:
    RevisionId: str
    Title: str
    Author: str
    Archived: bool
    Audiobook: bool
    Owner: User


class BookType(Enum):
    EBOOK = 1
    AUDIOBOOK = 2
    SUBSCRIPTION = 3


class NotAuthenticatedException(Exception):
    pass


class KoboException(Exception):
    pass


class Kobo:
    Affiliate = "Kobo"
    ApplicationVersion = "8.11.24971"
    DefaultPlatformId = "00000000-0000-0000-0000-000000004000"
    DisplayProfile = "Android"
    UserAgent = "Mozilla/5.0 (Linux; Android 6.0; Google Nexus 7 2013 Build/MRA58K; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/74.0.3729.186 Safari/537.36 KoboApp/8.40.2.29861 KoboPlatform Id/00000000-0000-0000-0000-000000004000 KoboAffiliate/Kobo KoboBuildFlavor/global"

    def __init__(self, user: User):
        self.InitializationSettings = {}
        self.Session = requests.session()
        self.Session.headers.update({
            "User-Agent": Kobo.UserAgent
        })
        self.user = user

    # PRIVATE METHODS

    # This could be added to the session but then we would need to add { "Authorization": None } headers to all other
    # functions that doesn't need authorization.
    def __GetHeaderWithAccessToken(self) -> dict:
        authorization = "Bearer " + self.user.AccessToken
        headers = {"Authorization": authorization}
        return headers

    def __RefreshAuthentication(self) -> None:
        headers = self.__GetHeaderWithAccessToken()

        postData = {
            "AppVersion": Kobo.ApplicationVersion,
            "ClientKey": base64.b64encode(Kobo.DefaultPlatformId.encode()).decode(),
            "PlatformId": Kobo.DefaultPlatformId,
            "RefreshToken": self.user.RefreshToken,
        }

        # The reauthentication hook is intentionally not set.
        response = self.Session.post(
            "https://storeapi.kobo.com/v1/auth/refresh", json=postData, headers=headers
        )
        debug_data("RefreshAuth", postData, response.text)
        response.raise_for_status()
        jsonResponse = response.json()

        if jsonResponse["TokenType"] != "Bearer":
            raise KoboException(
                "Authentication refresh returned with an unsupported token type: '%s'"
                % jsonResponse["TokenType"]
            )

        self.user.AccessToken = jsonResponse["AccessToken"]
        self.user.RefreshToken = jsonResponse["RefreshToken"]
        if not self.user.AreAuthenticationSettingsSet():
            raise KoboException("Authentication settings are not set after authentication refresh.")

        Globals.Settings.Save()

    # This could be added to the session too. See the comment at GetHeaderWithAccessToken.
    def __GetReauthenticationHook(self) -> dict:
        # The hook's workflow is based on this:
        # https://github.com/requests/toolbelt/blob/master/requests_toolbelt/auth/http_proxy_digest.py
        def ReauthenticationHook(r, *args, **kwargs):
            debug_data("Response", r.text)
            if r.status_code != requests.codes.unauthorized:  # 401
                return

            print("Refreshing expired authentication token...", file=sys.stderr)

            # Consume content and release the original connection to allow our new request to reuse the same one.
            r.content
            r.close()

            prep = r.request.copy()

            # Refresh the authentication token and use it.
            self.__RefreshAuthentication()
            headers = self.__GetHeaderWithAccessToken()
            prep.headers["Authorization"] = headers["Authorization"]

            # Don't retry to reauthenticate this request again.
            prep.deregister_hook("response", ReauthenticationHook)

            # Resend the failed request.
            _r = r.connection.send(prep, **kwargs)
            _r.history.append(r)
            _r.request = prep
            return _r

        return {"response": ReauthenticationHook}

    def __GetExtraLoginParameters(self) -> Tuple[str, str, str]:
        signInUrl = self.InitializationSettings["sign_in_page"]

        params = {
            "wsa": Kobo.Affiliate,
            "pwsav": Kobo.ApplicationVersion,
            "pwspid": Kobo.DefaultPlatformId,
            "pwsdid": self.user.DeviceId,
        }

        response = self.Session.get(signInUrl, params=params)
        response.raise_for_status()
        htmlResponse = response.text

        # The link can be found in the response ('<a class="kobo-link partner-option kobo"') but this will do for now.
        parsed = urllib.parse.urlparse(signInUrl)
        koboSignInUrl = parsed._replace(query=None, path="/ww/en/signin/signin").geturl()

        match = re.search(r"""\?workflowId=([^"]{36})""", htmlResponse)
        if match is None:
            raise KoboException(
                "Can't find the workflow ID in the login form. The page format might have changed."
            )
        workflowId = html.unescape(match.group(1))

        match = re.search(
            r"""<input name="__RequestVerificationToken" type="hidden" value="([^"]+)" />""",
            htmlResponse,
        )
        if match is None:
            raise KoboException(
                "Can't find the request verification token in the login form. The page format might have changed."
            )
        requestVerificationToken = html.unescape(match.group(1))

        return koboSignInUrl, workflowId, requestVerificationToken

    def __GetMyBookListPage(self, syncToken: str) -> Tuple[list, str]:
        url = self.InitializationSettings["library_sync"]
        headers = self.__GetHeaderWithAccessToken()
        hooks = self.__GetReauthenticationHook()

        if len(syncToken) > 0:
            headers["x-kobo-synctoken"] = syncToken

        debug_data("GetMyBookListPage")
        response = self.Session.get(url, headers=headers, hooks=hooks)
        response.raise_for_status()
        bookList = response.json()

        syncToken = ""
        syncResult = response.headers.get("x-kobo-sync")
        if syncResult == "continue":
            syncToken = response.headers.get("x-kobo-synctoken", "")

        return bookList, syncToken

    def __GetContentAccessBook(self, productId: str, displayProfile: str) -> dict:
        url = self.InitializationSettings["content_access_book"].replace("{ProductId}", productId)
        params = {"DisplayProfile": displayProfile}
        headers = self.__GetHeaderWithAccessToken()
        hooks = self.__GetReauthenticationHook()

        debug_data("GetContentAccessBook")
        response = self.Session.get(url, params=params, headers=headers, hooks=hooks)
        response.raise_for_status()
        jsonResponse = response.json()
        return jsonResponse

    @staticmethod
    def __GetContentKeys(contentAccessBookResponse: dict) -> Dict[str, str]:
        jsonContentKeys = contentAccessBookResponse.get("ContentKeys")
        if jsonContentKeys is None:
            return {}

        contentKeys = {}
        for contentKey in jsonContentKeys:
            contentKeys[contentKey["Name"]] = contentKey["Value"]
        return contentKeys

    @staticmethod
    def __getContentUrls(bookMetadata: dict) -> str:
        keys = bookMetadata.keys()
        jsonContentUrls = None
        if 'ContentUrls' in keys:
            jsonContentUrls = bookMetadata.get("ContentUrls")
        if 'DownloadUrls' in keys:
            jsonContentUrls = bookMetadata.get('DownloadUrls')
        return jsonContentUrls

    def __GetDownloadInfo(
        self, bookMetadata: dict, isAudiobook: bool, displayProfile: str = None
    ) -> Tuple[str, bool]:
        displayProfile = displayProfile or Kobo.DisplayProfile
        productId = Kobo.GetProductId(bookMetadata)

        if not isAudiobook:
            jsonResponse = self.__GetContentAccessBook(productId, displayProfile)
            jsonContentUrls = Kobo.__getContentUrls(jsonResponse)
        else:
            jsonContentUrls = Kobo.__getContentUrls(bookMetadata)

        if jsonContentUrls is None:
            raise KoboException(f"Download URL can't be found for product {productId}.")

        if len(jsonContentUrls) == 0:
            raise KoboException(
                f"Download URL list is empty for product '{productId}'. If this is an archived book then it must be unarchived first on the Kobo website (https://www.kobo.com/help/en-US/article/1799/restoring-deleted-books-or-magazines)."
            )

        for jsonContentUrl in jsonContentUrls:
            drm_keys = ['DrmType', 'DRMType']
            drm_types = ["KDRM", "AdobeDrm"]
            # will be empty (falsey) if the drm listed doesn't match one of the drm_types
            hasDrm = [
                jsonContentUrl.get(key)
                for key in drm_keys
                if (jsonContentUrl.get(key) in drm_types)
            ]

            download_keys = ['DownloadUrl', 'Url']
            for key in download_keys:
                download_url = jsonContentUrl.get(key, None)
                if download_url:
                    return download_url, hasDrm

        message = f"Download URL for supported formats can't be found for product '{productId}'.\n"
        message += "Available formats:"
        for jsonContentUrl in jsonContentUrls:
            message += f'\nDRMType: \'{jsonContentUrl["DRMType"]}\', UrlFormat: \'{jsonContentUrl["UrlFormat"]}\''
        raise KoboException(message)

    def __DownloadToFile(self, url, outputPath: str) -> None:
        response = self.Session.get(url, stream=True)
        response.raise_for_status()
        with open(outputPath, "wb") as f:
            for chunk in response.iter_content(chunk_size=1024 * 256):
                f.write(chunk)

    def __DownloadAudiobook(self, url, outputPath: str) -> None:
        response = self.Session.get(url)

        response.raise_for_status()
        if not os.path.isdir(outputPath):
            os.mkdir(outputPath)
        data = response.json()

        for item in data['Spine']:
            fileNum = int(item['Id']) + 1
            response = self.Session.get(item['Url'], stream=True)
            filePath = os.path.join(outputPath, str(fileNum) + '.' + item['FileExtension'])
            with open(filePath, "wb") as f:
                for chunk in response.iter_content(chunk_size=1024 * 256):
                    f.write(chunk)

    # PUBLIC METHODS:
    @staticmethod
    def GetProductId(bookMetadata: dict) -> str:
        revisionId = bookMetadata.get('RevisionId')
        Id = bookMetadata.get('Id')
        return revisionId or Id

    # The initial device authentication request for a non-logged in user doesn't require a user key, and the returned
    # user key can't be used for anything.
    def AuthenticateDevice(self, userKey: str = "") -> None:
        if len(self.user.DeviceId) == 0:
            self.user.DeviceId = str(uuid.uuid4())
            self.user.AccessToken = ""
            self.user.RefreshToken = ""

        postData = {
            "AffiliateName": Kobo.Affiliate,
            "AppVersion": Kobo.ApplicationVersion,
            "ClientKey": base64.b64encode(Kobo.DefaultPlatformId.encode()).decode(),
            "DeviceId": self.user.DeviceId,
            "PlatformId": Kobo.DefaultPlatformId,
        }

        if len(userKey) > 0:
            postData["UserKey"] = userKey

        response = self.Session.post("https://storeapi.kobo.com/v1/auth/device", json=postData)
        debug_data("AuthenticateDevice", response.text)
        response.raise_for_status()
        jsonResponse = response.json()

        if jsonResponse["TokenType"] != "Bearer":
            raise KoboException(
                "Device authentication returned with an unsupported token type: '%s'"
                % jsonResponse["TokenType"]
            )

        self.user.AccessToken = jsonResponse["AccessToken"]
        self.user.RefreshToken = jsonResponse["RefreshToken"]
        if not self.user.AreAuthenticationSettingsSet():
            raise KoboException("Authentication settings are not set after device authentication.")

        if len(userKey) > 0:
            self.user.UserKey = jsonResponse["UserKey"]

        Globals.Settings.Save()

    # Downloading archived books is not possible, the "content_access_book" API endpoint returns with empty ContentKeys
    # and ContentUrls for them.
    def Download(self, bookMetadata: dict, isAudiobook: bool, outputPath: str) -> None:

        downloadUrl, hasDrm = self.__GetDownloadInfo(bookMetadata, isAudiobook)
        revisionId = Kobo.GetProductId(bookMetadata)
        temporaryOutputPath = outputPath + ".downloading"

        try:
            if isAudiobook:
                self.__DownloadAudiobook(downloadUrl, outputPath)
            else:
                self.__DownloadToFile(downloadUrl, temporaryOutputPath)

            if hasDrm:
                if hasDrm[0] == 'AdobeDrm':
                    print(
                        "WARNING: Unable to parse the Adobe Digital Editions DRM. Saving it as an encrypted 'ade' file.",
                        "Try https://github.com/apprenticeharper/DeDRM_tools",
                    )
                    copyfile(temporaryOutputPath, outputPath + ".ade")
                else:
                    contentAccessBook = self.__GetContentAccessBook(revisionId, self.DisplayProfile)
                    contentKeys = Kobo.__GetContentKeys(contentAccessBook)
                    drmRemover = KoboDrmRemover(self.user.DeviceId, self.user.UserId)
                    drmRemover.RemoveDrm(temporaryOutputPath, outputPath, contentKeys)
                os.remove(temporaryOutputPath)
            else:
                if not isAudiobook:
                    os.rename(temporaryOutputPath, outputPath)
        except:
            if os.path.isfile(temporaryOutputPath):
                os.remove(temporaryOutputPath)
            if os.path.isfile(outputPath):
                os.remove(outputPath)

            raise

    # The "library_sync" name and the synchronization tokens make it somewhat suspicious that we should use
    # "library_items" instead to get the My Books list, but "library_items" gives back less info (even with the
    # embed=ProductMetadata query parameter set).
    def GetMyBookList(self) -> list:

        if not self.user.AreAuthenticationSettingsSet():
            raise NotAuthenticatedException(f'User {self.user.Email} is not authenticated')

        fullBookList = []
        syncToken = ""
        while True:
            bookList, syncToken = self.__GetMyBookListPage(syncToken)
            fullBookList += bookList
            if len(syncToken) == 0:
                break

        return fullBookList

    def GetMyWishList(self) -> list:
        items = []
        currentPageIndex = 0

        while True:
            url = self.InitializationSettings["user_wishlist"]
            headers = self.__GetHeaderWithAccessToken()
            hooks = self.__GetReauthenticationHook()

            params = {
                "PageIndex": currentPageIndex,
                "PageSize": 100,  # 100 is the default if PageSize is not specified.
            }

            debug_data("GetMyWishList")
            response = self.Session.get(url, params=params, headers=headers, hooks=hooks)
            response.raise_for_status()
            wishList = response.json()

            items.extend(wishList["Items"])

            currentPageIndex += 1
            if currentPageIndex >= wishList["TotalPageCount"]:
                break

        return items

    def GetBookInfo(self, productId: str) -> dict:
        audiobook_url = self.InitializationSettings["audiobook"].replace("{ProductId}", productId)
        ebook_url = self.InitializationSettings["book"].replace("{ProductId}", productId)
        headers = self.__GetHeaderWithAccessToken()
        hooks = self.__GetReauthenticationHook()
        debug_data("GetBookInfo")
        try:
            response = self.Session.get(ebook_url, headers=headers, hooks=hooks)
            response.raise_for_status()
        except requests.HTTPError as err:
            response = self.Session.get(audiobook_url, headers=headers, hooks=hooks)
            response.raise_for_status()
        jsonResponse = response.json()
        return jsonResponse

    def LoadInitializationSettings(self) -> None:
        """
        to be called when authentication has been done
        """
        headers = self.__GetHeaderWithAccessToken()
        hooks = self.__GetReauthenticationHook()
        debug_data("LoadInitializationSettings")
        response = self.Session.get(
            "https://storeapi.kobo.com/v1/initialization", headers=headers, hooks=hooks
        )
        try:
            response.raise_for_status()
            jsonResponse = response.json()
            self.InitializationSettings = jsonResponse["Resources"]
        except requests.HTTPError as err:
            print(response.reason, response.text)
            raise err

    def Login(self, email: str, password: str, captcha: str) -> None:
        (
            signInUrl,
            workflowId,
            requestVerificationToken,
        ) = self.__GetExtraLoginParameters()

        postData = {
            "LogInModel.WorkflowId": workflowId,
            "LogInModel.Provider": Kobo.Affiliate,
            "ReturnUrl": "",
            "__RequestVerificationToken": requestVerificationToken,
            "LogInModel.UserName": email,
            "LogInModel.Password": password,
            "g-recaptcha-response": captcha,
            "h-captcha-response": captcha,
        }

        response = self.Session.post(signInUrl, data=postData)
        debug_data("Login", response.text)
        response.raise_for_status()
        htmlResponse = response.text

        match = re.search(r"'(kobo://UserAuthenticated\?[^']+)';", htmlResponse)
        if match is None:
            soup = BeautifulSoup(htmlResponse, 'html.parser')
            errors = soup.find(class_='validation-summary-errors') or soup.find(
                class_='field-validation-error'
            )
            if errors:
                raise KoboException('Login Failure! ' + errors.text)
            else:
                with open('loginpage_error.html', 'w') as loginpagefile:
                    loginpagefile.write(htmlResponse)
                raise KoboException(
                    "Authenticated user URL can't be found. The page format might have changed!\n\n"
                    "The bad page has been written to file 'loginpage_error.html'.  \n"
                    "You should open an issue on GitHub and attach this file for help: https://github.com/subdavis/kobo-book-downloader/issues\n"
                    "Please be sure to remove any personally identifying information from the file."
                )

        url = match.group(1)
        parsed = urllib.parse.urlparse(url)
        parsedQueries = urllib.parse.parse_qs(parsed.query)
        self.user.UserId = parsedQueries["userId"][
            0
        ]  # We don't call self.Settings.Save here, AuthenticateDevice will do that if it succeeds.
        userKey = parsedQueries["userKey"][0]

        self.AuthenticateDevice(userKey)



import json
import os
import platform
from typing import List, TextIO, Tuple, Union

import click

from kobodl.globals import Globals
from kobodl.kobo import Book, BookType, Kobo, KoboException, NotAuthenticatedException
from kobodl.settings import User

SUPPORTED_BOOK_TYPES = [
    BookType.EBOOK,
    BookType.AUDIOBOOK,
]


def __GetBookAuthor(book: dict) -> str:
    contributors = book.get('ContributorRoles')

    authors = []
    for contributor in contributors:
        role = contributor.get('Role')
        if role == 'Author':
            authors.append(contributor['Name'])

    # Unfortunately the role field is not filled out in the data returned by the 'library_sync' endpoint, so we only
    # use the first author and hope for the best. Otherwise we would get non-main authors too. For example Christopher
    # Buckley beside Joseph Heller for the -- terrible -- novel Catch-22.
    if len(authors) == 0 and len(contributors) > 0:
        authors.append(contributors[0]['Name'])

    return ' & '.join(authors)


def __SanitizeString(string: str) -> str:
    result = ''
    for c in string:
        if c.isalnum() or ' ,;.!(){}[]#$\'-+@_'.find(c) >= 0:
            result += c

    result = result.strip(' .')
    if platform.system() == 'Windows':
        # Limit the length -- mostly because of Windows. It would be better to do it on the full path using MAX_PATH.
        result = result[:100]
    return result


def __MakeFileNameForBook(bookMetadata: dict, formatStr: str) -> str:
    '''filename without extension'''
    fileName = ''
    author = __SanitizeString(__GetBookAuthor(bookMetadata))
    title = __SanitizeString(bookMetadata['Title'])

    return formatStr.format_map(
        {
            **bookMetadata,
            'Author': author,
            'Title': title,
            # Append a portion of revisionId to prevent name collisions.
            'ShortRevisionId': bookMetadata['RevisionId'][:8],
        }
    )


def __GetBookMetadata(entitlement: dict) -> Tuple[dict, BookType]:
    keys = entitlement.keys()
    if 'BookMetadata' in keys:
        return entitlement['BookMetadata'], BookType.EBOOK
    if 'AudiobookMetadata' in keys:
        return entitlement['AudiobookMetadata'], BookType.AUDIOBOOK
    if 'BookSubscriptionEntitlement' in keys:
        return entitlement['BookSubscriptionEntitlement'], BookType.SUBSCRIPTION
    print(f'WARNING: unsupported object detected with contents {entitlement}')
    print('Please open an issue at https://github.com/subdavis/kobo-book-downloader/issues')
    return None, None


def __IsBookArchived(newEntitlement: dict) -> bool:
    keys = newEntitlement.keys()
    bookEntitlement: dict = {}
    if 'BookEntitlement' in keys:
        bookEntitlement = newEntitlement['BookEntitlement']
    if 'AudiobookEntitlement' in keys:
        bookEntitlement = newEntitlement['AudiobookEntitlement']
    return bookEntitlement.get('IsRemoved', False)


def __IsBookRead(newEntitlement: dict) -> bool:
    readingState = newEntitlement.get('ReadingState')
    if readingState is None:
        return False

    statusInfo = readingState.get('StatusInfo')
    if statusInfo is None:
        return False

    status = statusInfo.get('Status')
    return status == 'Finished'


def __GetBookList(kobo: Kobo, listAll: bool, exportFile: Union[TextIO, None]) -> list:
    bookList = kobo.GetMyBookList()
    rows = []

    if exportFile:
        exportFile.write(json.dumps(bookList, indent=2))

    for entitlement in bookList:
        newEntitlement = entitlement.get('NewEntitlement')
        if newEntitlement is None:
            continue

        bookEntitlement = newEntitlement.get('BookEntitlement')
        if bookEntitlement is not None:
            # Skip saved previews.
            if bookEntitlement.get('Accessibility') == 'Preview':
                continue

            # Skip refunded books.
            if bookEntitlement.get('IsLocked'):
                continue

        if (not listAll) and __IsBookRead(newEntitlement):
            continue

        bookMetadata, book_type = __GetBookMetadata(newEntitlement)

        if book_type is None:
            click.echo('Skipping book of unknown type')
            continue

        elif book_type in SUPPORTED_BOOK_TYPES:
            book = [
                bookMetadata['RevisionId'],
                bookMetadata['Title'],
                __GetBookAuthor(bookMetadata),
                __IsBookArchived(newEntitlement),
                book_type == BookType.AUDIOBOOK,
            ]
            rows.append(book)

    rows = sorted(rows, key=lambda columns: columns[1].lower())
    return rows


def ListBooks(users: List[User], listAll: bool, exportFile: Union[TextIO, None]) -> List[Book]:
    '''list all books currently in the account'''
    for user in users:
        kobo = Kobo(user)
        kobo.LoadInitializationSettings()
        rows = __GetBookList(kobo, listAll, exportFile)
        for columns in rows:
            yield Book(
                RevisionId=columns[0],
                Title=columns[1],
                Author=columns[2],
                Archived=columns[3],
                Audiobook=columns[4],
                Owner=user,
            )


def Login(user: User, password: str, captcha: str) -> None:
    '''perform device initialization and get token'''
    kobo = Kobo(user)
    kobo.AuthenticateDevice()
    kobo.LoadInitializationSettings()
    kobo.Login(user.Email, password, captcha)


def GetBookOrBooks(
    user: User,
    outputPath: str,
    formatStr: str = r'{Author} - {Title} {ShortRevisionId}',
    productId: str = '',
) -> Union[None, str]:
    """
    download 1 or all books to file
    returns output filepath if identifier is passed, otherwise returns None
    """
    outputPath = os.path.abspath(outputPath)
    kobo = Kobo(user)
    kobo.LoadInitializationSettings()

    # Must call GetBookList every time, even if you're only getting 1 book,
    # because it invokes a library sync endpoint.
    # This is the only known endpoint that returns
    # download URLs along with book metadata.
    bookList = kobo.GetMyBookList()

    for entitlement in bookList:
        newEntitlement = entitlement.get('NewEntitlement')
        if newEntitlement is None:
            continue

        bookMetadata, book_type = __GetBookMetadata(newEntitlement)
        if book_type is None:
            click.echo('Skipping book of unknown type')
            continue

        elif book_type == BookType.SUBSCRIPTION:
            click.echo('Skipping subscribtion entity')
            continue

        fileName = __MakeFileNameForBook(bookMetadata, formatStr)
        if book_type == BookType.EBOOK:
            # Audiobooks go in sub-directories
            # but epub files go directly in outputPath
            fileName += '.epub'
        outputFilePath = os.path.join(outputPath, fileName)

        if not productId and os.path.exists(outputFilePath):
            # when downloading ALL books, skip books we've downloaded before
            click.echo(f'Skipping already downloaded book {outputFilePath}')
            continue

        currentProductId = Kobo.GetProductId(bookMetadata)
        if productId and productId != currentProductId:
            # user only asked for a single title,
            # and this is not the book they want
            continue

        # Skip archived books.
        if __IsBookArchived(newEntitlement):
            click.echo(f'Skipping archived book {fileName}')
            continue

        try:
            click.echo(f'Downloading {currentProductId} to {outputFilePath}', err=True)
            kobo.Download(bookMetadata, book_type == BookType.AUDIOBOOK, outputFilePath)
        except KoboException as e:
            if productId:
                raise e
            else:
                click.echo(
                    (
                        f'Skipping failed download for {currentProductId}: {str(e)}'
                        '\n  -- Try downloading it as a single book to get the complete exception details'
                        ' and open an issue on the project GitHub page: https://github.com/subdavis/kobo-book-downloader/issues'
                    ),
                    err=True,
                )

        if productId:
            # TODO: support audiobook downloads from web
            return outputFilePath

    return None



import sys

import click

from kobodl.app import app
from kobodl.debug import debug_data
from kobodl.globals import Globals
from kobodl.settings import Settings


@click.group()
@click.option(
    '--fmt',
    type=click.STRING,
    default='simple',
    help='python-tabulate table format string',
)
@click.option(
    '--config',
    type=click.Path(dir_okay=False, file_okay=True, writable=True),
    help='path to kobodl.json config file',
)
@click.option(
    '--debug',
    is_flag=True,
    help="enable the debug log",
)
@click.version_option()
@click.pass_context
def cli(ctx, fmt, config, debug):
    Globals.Settings = Settings(config)
    Globals.Debug = debug
    ctx.obj = {
        'fmt': fmt,
        'debug': debug,
    }
    debug_data(sys.argv)


@click.command(name='serve', short_help='start an http server')
@click.option('-h', '--host', type=click.STRING)
@click.option('-p', '--port', type=click.INT)
@click.option('--debug', is_flag=True)
@click.option(
    '-o',
    '--output-dir',
    type=click.Path(file_okay=False, dir_okay=True, writable=True),
    default='kobo_downloads',
)
def serve(host, port, debug, output_dir):
    Globals.Debug = debug
    app.config['output_dir'] = output_dir
    app.run(host, port, debug)


cli.add_command(serve)

from kobodl.commands import book, user  # isort:skip noqa: F401 E402



import json.decoder
import os

from flask import Flask, abort, redirect, render_template, request, send_from_directory
from requests import Request, Session
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError

from kobodl import actions
from kobodl.globals import Globals
from kobodl.settings import User

app = Flask(__name__)


@app.route('/')
def index():
    return redirect('/user')


@app.route('/user', methods=['GET', 'POST'])
def users():
    error = None
    if request.method == 'POST':
        print(request.form)
        email = request.form.get('email')
        password = request.form.get('password')
        captcha = request.form.get('captcha')
        print(email, password, captcha)
        if email and password and captcha:
            user = User(Email=email)
            try:
                actions.Login(user, password, captcha)
                Globals.Settings.UserList.users.append(user)
                Globals.Settings.Save()
            except Exception as err:
                error = str(err)
        else:
            error = 'email, password, or captcha missing'
    users = Globals.Settings.UserList.users
    return render_template('users.j2', users=users, error=error)


@app.route('/user/<userid>/book', methods=['GET'])
def getUserBooks(userid, error=None, success=None):
    user = Globals.Settings.UserList.getUser(userid)
    if not user:
        abort(404)
    books = actions.ListBooks([user], False, None)
    calibre = Globals.Settings.UserList.calibre_web
    return render_template(
        'books.j2',
        books=books,
        calibre=calibre,
        error=error,
        success=success,
    )


@app.route('/user/<userid>/book/<productid>', methods=['GET'])
def downloadBook(userid, productid):
    user = Globals.Settings.UserList.getUser(userid)
    if not user:
        abort(404)
    outputDir = app.config.get('output_dir')
    os.makedirs(outputDir, exist_ok=True)
    # GetBookOrBooks always returns an absolute path
    outputFileName = actions.GetBookOrBooks(user, outputDir, productId=productid)
    absOutputDir, tail = os.path.split(outputFileName)
    # send_from_directory must be given an absolute path to avoid confusion
    # (relative paths are relative to root_path, not working dir)
    return send_from_directory(absOutputDir, tail, as_attachment=True, attachment_filename=tail)


@app.route('/user/<userid>/book/<productid>/send_to_calibre', methods=['POST'])
def sendToCalibre(userid, productid):
    user = Globals.Settings.UserList.getUser(userid)
    if not user:
        abort(404)
    outputDir = app.config.get('output_dir')
    os.makedirs(outputDir, exist_ok=True)
    # GetBookOrBooks always returns an absolute path
    outputFileName = actions.GetBookOrBooks(user, outputDir, productId=productid)
    req = Request(
        'POST',
        url=Globals.Settings.UserList.calibre_web.url,
        files={'btn-upload': open(outputFileName, 'rb')},
    )
    if Globals.Settings.UserList.calibre_web.username:
        req.auth = HTTPBasicAuth(
            Globals.Settings.UserList.calibre_web.username,
            Globals.Settings.UserList.calibre_web.password,
        )
    success = None
    try:
        resp = Session().send(req.prepare())
        resp.raise_for_status()
        success = f"New book uploaded to {resp.json()['location']}"
    except HTTPError as err:
        return getUserBooks(userid=userid, error=err)
    except json.decoder.JSONDecodeError:
        return getUserBooks(
            userid=userid, error="Could not decode response.  Check your CalibreWeb Credentials."
        )
    return getUserBooks(userid=userid, success=success)


@app.route('/book', methods=['GET'])
def books():
    userlist = Globals.Settings.UserList.users
    books = actions.ListBooks(userlist, False, None)
    calibre = Globals.Settings.UserList.calibre_web
    return render_template('books.j2', books=books, calibre=calibre)



import base64
import binascii
import hashlib
import zipfile
from typing import Dict

from Crypto.Cipher import AES
from Crypto.Util import Padding


# Based on obok.py by Physisticated.
class KoboDrmRemover:
    def __init__(self, deviceId: str, userId: str):
        self.DeviceIdUserIdKey = KoboDrmRemover.__MakeDeviceIdUserIdKey(deviceId, userId)

    @staticmethod
    def __MakeDeviceIdUserIdKey(deviceId: str, userId: str) -> bytes:
        deviceIdUserId = (deviceId + userId).encode()
        key = hashlib.sha256(deviceIdUserId).hexdigest()
        return binascii.a2b_hex(key[32:])

    def __DecryptContents(self, contents: bytes, contentKeyBase64: str) -> bytes:
        contentKey = base64.b64decode(contentKeyBase64)
        keyAes = AES.new(self.DeviceIdUserIdKey, AES.MODE_ECB)
        decryptedContentKey = keyAes.decrypt(contentKey)

        contentAes = AES.new(decryptedContentKey, AES.MODE_ECB)
        decryptedContents = contentAes.decrypt(contents)
        return Padding.unpad(decryptedContents, AES.block_size, "pkcs7")

    def RemoveDrm(self, inputPath: str, outputPath: str, contentKeys: Dict[str, str]) -> None:
        with zipfile.ZipFile(inputPath, "r") as inputZip:
            with zipfile.ZipFile(outputPath, "w", zipfile.ZIP_DEFLATED) as outputZip:
                for filename in inputZip.namelist():
                    contents = inputZip.read(filename)
                    contentKeyBase64 = contentKeys.get(filename, None)
                    if contentKeyBase64 is not None:
                        contents = self.__DecryptContents(contents, contentKeyBase64)
                    outputZip.writestr(filename, contents)



import dataclasses
import os
from typing import List, Union

from dataclasses_json import dataclass_json


@dataclass_json
@dataclasses.dataclass
class CalibreWeb:
    enabled: bool = False
    url: str = "http://localhost:8083"
    username: str = ""
    password: str = ""


@dataclass_json
@dataclasses.dataclass
class User:
    Email: str
    DeviceId: str = ""
    AccessToken: str = ""
    RefreshToken: str = ""
    UserId: str = ""
    UserKey: str = ""

    def AreAuthenticationSettingsSet(self) -> bool:
        return len(self.DeviceId) > 0 and len(self.AccessToken) > 0 and len(self.RefreshToken) > 0

    def IsLoggedIn(self) -> bool:
        return len(self.UserId) > 0 and len(self.UserKey) > 0


@dataclass_json
@dataclasses.dataclass
class UserList:
    users: List[User] = dataclasses.field(default_factory=list)
    calibre_web: CalibreWeb = dataclasses.field(default_factory=CalibreWeb)

    def getUser(self, identifier: str) -> Union[User, None]:
        for user in self.users:
            if (
                user.Email == identifier
                or user.UserKey == identifier
                or user.DeviceId == identifier
            ):
                return user
        return None

    def removeUser(self, identifier: str) -> Union[User, None]:
        """returns the removed user"""
        user = self.getUser(identifier)
        if user:
            i = self.users.index(user)
            return self.users.pop(i)
        return None


class Settings:
    def __init__(self, configpath=None):
        self.SettingsFilePath = configpath or Settings.__GetCacheFilePath()
        self.UserList = self.Load()

    def Load(self) -> UserList:
        if not os.path.isfile(self.SettingsFilePath):
            return UserList()
        with open(self.SettingsFilePath, "r") as f:
            jsonText = f.read()
            return UserList.from_json(jsonText)

    def Save(self) -> None:
        with open(self.SettingsFilePath, "w") as f:
            f.write(self.UserList.to_json(indent=4))

    @staticmethod
    def __GetCacheFilePath() -> str:
        cacheHome = os.environ.get("XDG_CONFIG_HOME")
        if (cacheHome is None) or (not os.path.isdir(cacheHome)):
            home = os.path.expanduser("~")
            cacheHome = os.path.join(home, ".config")
            if not os.path.isdir(cacheHome):
                cacheHome = home

        return os.path.join(cacheHome, "kobodl.json")



import os
from pathlib import Path
from typing import List

import click
from tabulate import tabulate

from kobodl import actions, cli
from kobodl.globals import Globals


def decorators(book):
    append = ''
    if book.Audiobook:
        append += ' (🎧 Audiobook)'
    if book.Archived:
        append += ' (🗄️ Archived)'
    return append


@click.group(name='book', short_help='list and download books')
def book():
    pass


@book.command(name='get', short_help='download book')
@click.option(
    '-u',
    '--user',
    type=click.STRING,
    help='Required when multiple accounts exist. Use either Email or UserKey',
)
@click.option(
    '-o',
    '--output-dir',
    type=click.Path(file_okay=False, dir_okay=True, writable=True),
    default='kobo_downloads',
    help='default: kobo_downloads',
)
@click.option('-a', '--get-all', is_flag=True)
@click.option(
    '-f',
    '--format-str',
    type=click.STRING,
    default=r'{Author} - {Title} {ShortRevisionId}',
    help=r"default: '{Author} - {Title} {ShortRevisionId}'",
)
@click.argument('product-id', nargs=-1, type=click.STRING)
@click.pass_obj
def get(ctx, user, output_dir: Path, get_all: bool, format_str: str, product_id: List[str]):
    if len(Globals.Settings.UserList.users) == 0:
        click.echo('error: no users found.  Did you `kobodl user add`?', err=True)
        exit(1)

    if not user:
        if len(Globals.Settings.UserList.users) > 1:
            click.echo('error: must provide --user option when more than 1 user exists.')
            exit(1)
        # Exactly 1 user account exists
        usercls = Globals.Settings.UserList.users[0]
    else:
        # A user was passed
        usercls = Globals.Settings.UserList.getUser(user)
        if not usercls:
            click.echo(f'error: could not find user with name or id {user}')
            exit(1)

    if get_all and len(product_id):
        click.echo(
            'error: cannot pass product IDs when --get-all is used. Use one or the other.',
            err=True,
        )
        exit(1)
    if not get_all and len(product_id) == 0:
        click.echo('error: must pass at least one Product ID, or use --get-all', err=True)
        exit(1)

    os.makedirs(output_dir, exist_ok=True)
    if get_all:
        actions.GetBookOrBooks(usercls, output_dir, formatStr=format_str)
    else:
        for pid in product_id:
            actions.GetBookOrBooks(usercls, output_dir, formatStr=format_str, productId=pid)


@book.command(name='list', help='list books')
@click.option(
    '-u',
    '--user',
    type=click.STRING,
    required=False,
    help='Limit list to a single user. Use either Email or UserKey',
)
@click.option('--read', is_flag=True, help='include books marked as read')
@click.option(
    '--export-library',
    type=click.File(mode='w'),
    help='filepath to write raw JSON library data to.',
)
@click.pass_obj
def list(ctx, user, read, export_library):
    userlist = Globals.Settings.UserList.users
    if user:
        userlist = [Globals.Settings.UserList.getUser(user)]
    books = actions.ListBooks(userlist, read, export_library)
    headers = ['Title', 'Author', 'RevisionId', 'Owner']
    data = sorted(
        [
            (
                book.Title + decorators(book),
                book.Author,
                book.RevisionId,
                book.Owner.Email,
            )
            for book in books
        ]
    )
    click.echo(tabulate(data, headers, tablefmt=ctx['fmt']))


cli.add_command(book)



from datetime import datetime

from kobodl.globals import Globals


def debug_data(*args):
    if Globals.Debug:
        with open('./debug.log', 'a', encoding='utf-8') as debuglog:
            debuglog.write(str(datetime.now()))
            debuglog.write('\n')
            for stringable in args:
                debuglog.write(str(stringable))
                debuglog.write('\n')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kobodl', 'kobodl.commands']

package_data = \
{'': ['*'], 'kobodl': ['templates/*']}

install_requires = \
['beautifulsoup4<5.0.0',
 'click<8',
 'dataclasses-json<0.6.0',
 'dataclasses<1.0.0',
 'flask>=1.1.0,<2.0.0',
 'pycryptodome<4',
 'pyperclip>=1.8.2,<2.0.0',
 'requests>=2.25,<3.0',
 'tabulate<0.9.0']

entry_points = \
{'console_scripts': ['kobodl = kobodl:cli']}

setup_kwargs = {
    'name': 'kobodl',
    'version': '0.9.1',
    'description': 'Kobo Book Downloader',
    'long_description': '## Alternatives to kobodl\n\nI used to prefer `kobodl` because it\'s **standalone**, which means you don\'t need other proprietary software like Adobe Digial Editions or Kindle for PC (that I can\'t use on Linux). However, I have since discoverd a way to do this with [Calibre](https://github.com/kovidgoyal/calibre) and 2 plugins:\n\n* [Leseratte10/acsm-calibre-plugin](https://github.com/Leseratte10/acsm-calibre-plugin) - A plugin that can read Adobe Digital Editions files that Kobo web download produces.\n* [noDRM/DeDRM_tools](https://github.com/noDRM/DeDRM_tools) - The popular DRM removal plugin.\n\nNow you can just download the `.acm` file from your book list on Kobo.com and load it into Calibre desktop!\n\nIt doesn\'t work with audiobooks and is a little harder to set up, but I think it\'s overall a better solution and I am not using kobodl personally much anymore. If anyone is interested in becoming a new maintainer of kobodl please let me know by opening a [discussion](https://github.com/subdavis/kobo-book-downloader/discussions).  I will still try to keep it functioning as long as I can, and bug reports are still appreciated.\n\n---\n\n![kobodl logo](docs/kobodl.png)\n\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/subdavis/kobo-book-downloader/Publish%20Docker%20Container?style=for-the-badge)\n![GitHub release (latest by date)](https://img.shields.io/github/v/release/subdavis/kobo-book-downloader?style=for-the-badge)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/kobodl?style=for-the-badge)\n![PyPI - License](https://img.shields.io/pypi/l/kobodl?style=for-the-badge)\n![PyPI](https://img.shields.io/pypi/v/kobodl?style=for-the-badge)\n\n# kobodl\n\nThis is a hard fork of [kobo-book-downloader](https://github.com/TnS-hun/kobo-book-downloader), a command line tool to download and remove Digital Rights Management (DRM) protection from media legally purchased from [Rakuten Kobo](https://www.kobo.com/). The resulting [EPUB](https://en.wikipedia.org/wiki/EPUB) files can be read with, amongst others, [KOReader](https://github.com/koreader/koreader).\n\n> **NOTE:** You must have a kobo email login for this tool to work (you can\'t use an external provider like Google or Facebook). However, you can create a NEW kobo account and link it with your existing account on the user profile page. Go to `My Account -> Account Settings` to link your new kobo login.\n\n## Features\n\nkobodl preserves the features from [TnS-hun/kobo-book-downloader](https://github.com/TnS-hun/kobo-book-downloader).\n\n* stand-alone; no need to run other software or pre-download through an e-reader.\n* downloads `.epub` formatted books\n\nIt adds several new features.\n\n* **audiobook support**; command-line only for now.\n  * Use `kobodl book get`. There will not be a download button in the webpage for audiobooks because they consist of many large files.\n* **multi-user support**; fetch books for multiple accounts.\n* **web interface**; adds new browser gui (with flask)\n* **CalibreWeb Integration**; send books directly to CalibreWeb from the web interface.\n* [docker image](https://github.com/subdavis/kobodl/pkgs/container/kobodl)\n* [pypi package](https://pypi.org/project/kobodl/)\n* [pyinstaller bundles](https://github.com/subdavis/kobo-book-downloader/releases/latest)\n\n## Web UI\n\nWebUI provides most of the same functions of the CLI. It was added to allow other members of a household to add their accounts to kobodl and access their books without having to set up python.\n\n### User page\n\n![Example of User page](docs/webss.png)\n\n### Book list page\n\n![Example of book list page](docs/books3.png)\n\n## Installation\n\n### pipx\n\n``` bash\npipx install kobodl\n```\n\n### pip\n\n``` bash\npip install kobodl\n```\n\n### Pre-built bundles\n\nNo python installation necessary.  Simply download the appropriate executable from [the latest release assets](https://github.com/subdavis/kobo-book-downloader/releases/latest) and run it from the command line.  Pre-built bundles are CLI-only (no web server) so use a different install option if you want that feature.\n\n``` bash\n# Linux\nwget https://github.com/subdavis/kobo-book-downloader/releases/latest/download/kobodl-ubuntu\nchmod +x kobodl-ubuntu-latest\n./kobodl-ubuntu-latest\n```\n\n``` bash\n# MacOS (Catalina 10.15 or newer required.  For older versions, use the pip or docker install option)\nwget https://github.com/subdavis/kobo-book-downloader/releases/latest/download/kobodl-macos\nchmod +x kobodl-macos-latest\n./kobodl-macos-latest\n```\n\n``` powershell\n# Windows Powershell example\nwget "https://github.com/subdavis/kobo-book-downloader/releases/latest/download/kobodl-windows.exe" -outfile "kobodl.exe"\n./kobodl.exe\n```\n\n### docker\n\n> *Note*: for rootless docker installations (uncommon), omit the `--user` argument.\n\n``` bash\n# list users\ndocker run --rm -it --user $(id -u):$(id -g) \\\n  -v ${HOME}/.config:/home/config \\\n  ghcr.io/subdavis/kobodl \\\n  --config /home/config/kobodl.json user list\n\n# run http server\ndocker run --rm -it --user $(id -u):$(id -g) \\\n  -p 5000:5000 \\\n  -v ${HOME}/.config:/home/config \\\n  -v ${PWD}:/home/downloads \\\n  ghcr.io/subdavis/kobodl \\\n  --config /home/config/kobodl.json \\\n  serve \\\n  --host 0.0.0.0 \\\n  --output-dir /home/downloads/kobodl_downloads\n```\n\n[Also see the **docker-compose** example file.](./docker-compose.yml)\n\n## Usage\n\nGeneral usage documentation.\n\n> **Note**: These are commands you type into a shell prompt like Terminal (Ubuntu, MacOS) or Powershell or CMD (Windows).  You may need to replace `kobodl` with `./kobodl.exe`, `./kobodl-macos-latest`, or something else, depending on which installation method you chose.\n\n``` bash\n# Get started by adding one or more users\n# See `Getting a reCAPTCHA code` below for more help\nkobodl user add\n\n# List users\nkobodl user list\n\n# Remove a user\nkobodl user rm email@domain.com\n\n# List books\nkobodl book list\n\n# List books for a single user\nkobodl book list --user email@domain.com\n\n# List all books, including those marked as read\nkobodl book list --read\n\n# Show book list help\nkobodl book list --help\n\n# Download a single book with default options when only 1 user exists\n# default output directory is `./kobo_downloads` \nkobodl book get c1db3f5c-82da-4dda-9d81-fa718d5d1d16\n\n# Download a single book with advanced options\nkobodl book get \\\n  --user email@domain.com \\\n  --output-dir /path/to/download_directory \\\n  --format-str \'{Title}\' \\\n  c1db3f5c-82da-4dda-9d81-fa718d5d1d16\n\n# Download ALL books with default options when only 1 user exists\nkobodl book get --get-all\n\n# Download ALL books with advanced options\nkobodl book get \\\n  --user email@domain.com \\\n  --output-dir /path/to/download_directory \\\n  --format-str \'{Title}\' \\\n  --get-all\n```\n\nRunning the web UI\n\n``` bash\nkobodl serve\n * Serving Flask app "kobodl.app" (lazy loading)\n * Environment: production\n   WARNING: This is a development server. Do not use it in a production deployment.\n   Use a production WSGI server instead.\n * Debug mode: off\n * Running on http://127.0.0.1:5000/ (Press CTRL+C to quit)\n```\n\nGlobal options\n\n``` bash\n# argument format\nkobodl [OPTIONS] COMMAND [ARGS]...\n\n# set python tabulate formatting style.\nkobodl --fmt "pretty" COMMAND [ARGS]...\n\n# set config path if different than ~/.config/kobodl.json\nkobodl --config /path/to/kobodl.json COMMAND [ARGS]...\n\n# get version\nkobodl --version\n\n# enable debugging, prints to "debug.log"\nkobodl --debug [OPTIONS] COMMAND [ARGS]...\n```\n\n## Getting a reCAPTCHA code\n\nAdding a user requires a bit of hackery to get a reCAPTCHA code from Kobo\'s website. This GIF helps to explain how to do that.\n\n![Gif explaining how to get reCAPTHCA](docs/captcha.gif)\n\n## CalibreWeb Integration: **Experimental**\n\nKobodl\'s web interface can send books directly to your calibre web server.  Only supported from web client.\n\nYou must add your CalibreWeb server and user details directly in kobodl.json.  Make sure to include `/upload` in the url.  If you don\'t have CalibreWeb authentication on, leave these fields blank (`""`).\n\n``` json\n{\n    "users": [ /* ... */ ],\n    "calibre_web": {\n        "enabled": true,\n        "url": "https://calibre.domain.com/upload",\n        "username": "CHANGEME",\n        "password": "CHANGEME"\n    }\n}\n```\n\n> **NOTE**: you may need to add escape characters if your password contains a `\\`, `"`, or others that require JSON escaping.\n\n## Troubleshooting\n\n> Some of my books are missing!\n\nTry `kobodl book list --read` to show all "finished" and "archived" books.  You can manage your book status on [the library page](https://kobo.com/library).  Try changing the status using the "..." button.\n\n> I see a mesage about "skipping _____" when I download all.\n\nTry to download the book individually using `kobodl book get <revision-id>`, replacing `revision-id` with the UUID from the list table.\n\n> I can\'t log in.  I get a message saying "The page format might have changed"\n\nThis happens from time to time, maybe once or twice a year.  Kobo changes their login page and makes it hard for the tool to parse out the necessary information.  Please open an issue.\n\n> Something else is going wrong!\n\nTry enabling debugging.  Run `kobodl --debug book get` (for example), which will dump a lot of data into a file called `debug.log`. Email me this file. Do not post it in public on an issue because it will contain information about your account.  My email address can be found on my [github profile page](https://github.com/subdavis).\n\n## Development\n\nThis project uses [Python Poetry](https://python-poetry.org/).\n\n```bash\npip install poetry\ngit clone https://github.com/subdavis/kobo-book-downloader\npoetry install\n\n# Run command line app\npoetry run kobodl\n\n# Run linting\npoetry run tox -e lint\n\n# Run standalone bundle generation\npoetry run tox -e buildcli\n\n# Run type checks\npoetry run tox -e type\n```\n\n## Notes\n\nkobo-book-downloader will prompt for your [Kobo](https://www.kobo.com/) e-mail address and password. Once it has successfully logged in, it won\'t ask for them again. Your password will not be stored on disk; Kobodl uses access tokens after the initial login.\n\nCredit recursively to [kobo-book-downloader](https://github.com/TnS-hun/kobo-book-downloader) and the projects that lead to it.\n\n## FAQ\n\n**How does this work?**\n\nkobodl works by pretending to be an Android Kobo e-reader.  It initializes a device, fetches your library, and downloads books as a "fake" Android app.\n\n**Why does this download KEPUB formatted books?**\n\nKobo has different formats that it serves to different platforms.  For example, Desktop users generally get `EPUB3` books with `AdobeDrm` DRM.  Android users typically get `KEPUB` books with `KDRM` DRM, which is fairly easy to remove, so that\'s what you get when you use this tool.\n\n**Is this tool safe and okay to use?**\n\nI\'m not a lawyer, and the discussion below is strictly academic.\n\nThe author(s) of `kobodl` don\'t collect any information about you or your account aside from what is made available through metrics from GitHub, PyPi, Docker Hub, etc.  See `LICENSE.md` for further info.\n\nKobo would probably claim that this tool violates its [Terms of Use](https://authorize.kobo.com/terms/termsofuse) but I\'m not able to conclusively determine that it does so.  Some relevant sections are reproduced here.\n\n> The download of, and access to any Digital Content is available only to Customers and is intended only for such Customers’ personal and non-commercial use. Any other use of Digital Content downloaded or accessed from the Service is strictly prohibited\n\nThis tool should only be used to download books for personal use.\n\n> You may not obscure or misrepresent your geographical location, forge headers, use proxies, use IP spoofing or otherwise manipulate identifiers in order to disguise the origin of any message or transmittal you send on or through the Service. You may not pretend that you are, or that you represent, someone else, or impersonate any other individual or entity.\n\nThis might be a violation.  This client announces itself to Kobo servers as an Android device, which can safely be construed as "manipulating identifiers", but whether or not the purpose is to "disguise the origin" is unclear.\n\n> Kobo may also take steps to prevent fraud, such as restricting the number of titles that may be accessed at one time, and monitoring Customer accounts for any activity that may violate these Terms. If Kobo discovers any type of fraud, Kobo reserves the right to take enforcement action including the termination or suspension of a User’s account.\n\nIn other words, you could have your account suspended for using `kobodl`. **Please open an issue on the issue tracker if this happens to you.**\n',
    'author': 'Brandon Davis',
    'author_email': 'kobodl@subdavis.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/subdavis/kobo-book-downloader.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<3.10',
}


setup(**setup_kwargs)
